from typing import Optional, Tuple

import rich_click as click

from hledger_args.output_result import output_report

from .batch_args import BatchArgs
from .inter_args import InteractiveArgs

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

click.rich_click.USE_MARKDOWN = True
click.rich_click.SHOW_ARGUMENTS = True
click.rich_click.GROUP_ARGUMENTS_OPTIONS = True
click.rich_click.MAX_WIDTH = 80
click.rich_click.SHOW_METAVARS_COLUMN = False
click.rich_click.APPEND_METAVARS_HELP = True
click.rich_click.STYLE_ERRORS_SUGGESTION = "magenta italic"
click.rich_click.ERRORS_SUGGESTION = (
    "Try running the '--help' flag for more information."
)
click.rich_click.ERRORS_EPILOGUE = "To find out more, visit [link=https://github.com/edkedk99/hledger-args]https://github.com/edkedk99/hledger-args[/link]"
click.rich_click.STYLE_OPTIONS_TABLE_LEADING = 1
click.rich_click.STYLE_OPTIONS_TABLE_BOX = "SIMPLE"
click.rich_click.STYLE_OPTIONS_PANEL_BORDER = "dim"  # Possibly conceal


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-f",
    "--file",
    type=click.STRING,
    required=True,
    help="Inform the journal file path",
)
@click.option(
    "-i",
    "--interactive",
    is_flag=True,
    required=False,
    help="Run in interactive mode by answering the prompts. [NAME] and [EXTRA_HLEDGER_OPTIONS] are not used in this mode.",
)
@click.option(
    "-d",
    "--pdf-dir",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
    ),
    required=False,
    help="Save the report to a folder named according to the current date under the specified directory",
)
@click.option(
    "-o",
    "--pdf-file",
    type=click.Path(file_okay=True, dir_okay=False),
    required=False,
    help="output the report to the specified file in pdf",
)
@click.argument("name", type=click.STRING, required=False)
@click.argument("extra_hledger_options", nargs=-1)
def cli(
    file: str,
    interactive: bool,
    name: str,
    extra_hledger_options: Tuple[str, ...],
    pdf_dir: Optional[str],
    pdf_file: Optional[str],
):
    """
     ---

     **NAME**: Command name to run saved in the journal sub directives. Not available in Interactive mode

    **EXTRA_HLEDGER_OPTIONS**: Extra options to send to hledger command. Not available in Interactive mode. The name **"all"** is special and output all the *no interactive* commands.

     ---

     In basic usage, this package is a replacement for [hledger argument file](https://hledger.org/1.29/hledger.html#command-arguments) using custom directives inside the journal file, instead of referencing to an argument file

     **Interactive Mode**: Instead of giving the desired command thru a command-line argument, choose it by selecting from a menu using the flag *--interactive*

     **Placeholder Command Substitution**: In Interactive Mode, a command can use placeholders by putting them between *curly braces* and additional prompts wil ask for the value and do the proper substitution

     **Special Placeholders**: *{account}, {payee}, {tag}, {tag_name}, {months}, {type}, {cur}* are special placeholders that offers autocomplete with fuzzy search using data from the journal. See the [README](https://github.com/edkedk99/hledger-args) for explanation on each of them

     **Sub directive format**:

     ```text
     #+args [command name]:[hledger options]
     #+args [other command_name]:[other hledger options]
     ```

     **Example**:

     ```text
     #+args buy_aapl:bal desc:\"Buy AAPL\"
     #+args aapl_cur:bal desc:\"Buy AAPL\" cur:{commodity}
    ```
    """

    if interactive:
        args = InteractiveArgs((file,))
    else:
        args = BatchArgs((file,), name, extra_hledger_options)

    output_report(args, pdf_dir, pdf_file)



import re
import shlex
import subprocess
import sys
from typing import List, Optional, Tuple

from .options import HledgerVars


def get_files_comm(file_path: Tuple[str, ...]) -> List[str]:
    files = []
    for file in file_path:
        files = [*files, "-f", file]
    return files


class BaseArgs:
    NAMESPACE = "args"

    def __init__(self, files: Tuple[str, ...]) -> None:
        self.files = files

        vars = HledgerVars(files)
        self.args = vars.get_namespace_vars(self.NAMESPACE)
        self.names = set(list(self.args.keys()))
        self.has_ask = {
            name for name, var in self.args.items() if re.search(r"\{(.*?)\}", var)
        }
        self.no_ask = self.names.difference(self.has_ask)

        self.files_comm = get_files_comm(files)

    def run_args(self, options: str, extra: Optional[Tuple[str, ...]] = None):
        options_list = shlex.split(options)
        if extra:
            options_list = [*options_list, *extra]

        base_comm = ["hledger", *self.files_comm, *options_list]
        proc = subprocess.run(base_comm, capture_output=True, check=True)
        report = proc.stdout.decode("utf8")

        base_comm_str = shlex.join(base_comm)
        print(f"stderr: {base_comm_str}\n", file=sys.stderr)
        return report



from datetime import datetime
from pathlib import Path
from typing import Optional, Union

from fpdf import FPDF

from .batch_args import BatchArgs
from .inter_args import InteractiveArgs


def create_pdf(header: str, text: str, output_path: Path):
    pdf = FPDF()
    pdf.add_page(orientation="L")
    pdf.set_font("Courier", "", 12)

    pdf.multi_cell(0, None, header, ln=True)
    lines = text.split("\n")

    for line in lines:
        if pdf.get_y() > pdf.h - 20:
            pdf.add_page()
        pdf.cell(0, None, line, ln=True)

    pdf.output(str(output_path))


def output_report(
    args: Union[BatchArgs, InteractiveArgs],
    pdf_dir: Optional[str],
    pdf_file: Optional[str],
):
    if pdf_file:
        file_dest = Path(pdf_file)
    elif pdf_dir:
        today = datetime.today().strftime("%Y%m%d")
        file_dir = Path(pdf_dir).joinpath(today)
        file_dir.mkdir(parents=True, exist_ok=True)
        name = args.name or "empty"
        file_dest = file_dir.joinpath(f"{name}.pdf")
    else:
        print(args.report)
        return

    header = f"""Report: {args.name}
_________________

"""

    create_pdf(header, args.report, file_dest)



import re
import subprocess
from datetime import datetime, timedelta
from typing import List, Optional, Tuple

import questionary
from prompt_toolkit.shortcuts import CompleteStyle

from .base_args import BaseArgs


def val_date(date: str):
    try:
        datetime.strptime(date, "%Y-%m-%d")
        return True
    except ValueError:
        return "Invalid date format"


def val_existing(choices: List[str], value: str):
    if value in choices:
        return True
    else:
        return "Only existing values are accepted"


def custom_autocomplete(name: str, choices: List[str]):
    question = questionary.autocomplete(
        f"{name} (TAB to autocomplete)",
        choices=choices,
        validate=lambda value: val_existing(choices, value),
        ignore_case=True,
        match_middle=True,
        style=questionary.Style([("answer", "fg:#f71b07")]),
        complete_style=CompleteStyle.MULTI_COLUMN,
    )
    return question


class InteractiveArgs(BaseArgs):
    def __init__(self, files: Tuple[str, ...]) -> None:
        super().__init__(files)
        self.name = self.menu()
        self.report = self.get_report()

    def ask_placeholder(self, placeholder: str) -> str:
        if placeholder == "account":
            choices = self.get_hledger_lines(["accounts"])
            answer = custom_autocomplete(placeholder, choices).ask()
        elif placeholder == "tag":
            tags = self.get_hledger_lines(["tags"])
            tag = custom_autocomplete(placeholder, choices=tags).ask()

            tag_values = self.get_hledger_lines(["tags", tag, "--values"])
            value = custom_autocomplete(
                f"Value for tag {tag}", choices=tag_values
            ).ask()
            answer = f'"tag:{tag}={value}"'
        elif placeholder.startswith("tag_"):
            tag_name = placeholder.split("_", 1)[1]
            tag_values = self.get_hledger_lines(["tags", tag_name, "--values"])
            value = custom_autocomplete(
                f"Value for tag {tag_name}", choices=tag_values
            ).ask()
            answer = f'"tag:{tag_name}={value}"'
        elif placeholder == "months":
            initial: str = questionary.text(
                "Initial", instruction="YYYY-MM-DD", validate=val_date
            ).ask()
            final = questionary.text(
                "Final (inclusive)", instruction="YYYY-MM-DD", validate=val_date
            ).ask()
            next_final_date = datetime.strptime(final, "%Y-%m-%d") + timedelta(days=1)
            next_final = next_final_date.strftime("%Y-%m-%d")
            answer = f"--begin {initial} --end {next_final}"
        elif placeholder == "payee":
            choices = self.get_hledger_lines(["payees"])
            payee = custom_autocomplete(placeholder, choices).ask()
            answer = f'"payee:{payee}"'
        elif placeholder == "cur":
            choices = self.get_hledger_lines(["commodities"])
            commodity = custom_autocomplete(placeholder, choices).ask()
            answer = f'"cur:{commodity}"'
        elif placeholder == "type":
            types = dict(
                Asset="A",
                Liability="L",
                Equity="E",
                Revenue="R",
                Expense="X",
                Cash="C",
                Conversion="V",
            )
            choices = list(types.keys())
            _type = questionary.select("Account Type", choices=choices).ask()
            type_code = types[_type]
            answer = f'"type:{type_code}"'

        else:
            answer = questionary.text(placeholder).ask()

        return answer

    def substitute(self, match: re.Match):
        placeholder: Optional[str] = match.group(1)
        if placeholder:
            question_desc = placeholder.replace("{", "").replace("}", "")
            answer = self.ask_placeholder(question_desc)
            return answer
        else:
            return ""

    def replace_options(self, options: str):
        result = re.sub(r"\{(.*?)\}", self.substitute, options)
        return result

    def get_hledger_lines(self, cmds: List[str]):
        base_comm = ["hledger", *self.files_comm, *cmds]
        proc = subprocess.run(base_comm, capture_output=True, check=True)
        report = proc.stdout.decode("utf8")
        report_list = [line for line in report.split("\n") if line != ""]
        return report_list

    def menu(self):
        name: str = questionary.select(
            "Choose report",
            choices=list(self.names),
            use_shortcuts=True,
            use_indicator=False,
            show_selected=False,
        ).ask()
        return name

    def get_report(self):
        options_str = self.args[self.name]
        replaced = self.replace_options(options_str)
        return self.run_args(replaced)



from dataclasses import dataclass
from typing import Dict, Optional, Tuple

NAMESPACE = "hledger-lots"
KEYS = {"avg_cost", "check", "no_desc"}


@dataclass
class Options:
    avg_cost: bool
    check: bool
    no_desc: str


class OptionError(BaseException):
    TEMPLATE = """Copy/Paste the text below to the journal, changing values according to the need

#+hledger-lots avg_cost:false, check:true
#+hledger-lots no_desc:
    """

    def __init__(self, message: str) -> None:
        self.message = f"\n\n{message}\n\n{self.TEMPLATE}"
        super().__init__(self.message)


class HledgerVars:
    NAMESPACE_START = "#+"

    def __init__(self, files: Tuple[str, ...]):
        self.files = files
        self.vars: Dict[str, str] = {}

    def get_var_tuple(self, var_str: str):
        var_key_value = var_str.split(":", 1)
        if len(var_key_value) == 0:
            return None

        key = var_key_value[0].strip(" \n")
        value = var_key_value[1].strip()
        result = (key, value)
        return result

    def get_row_vars(self, row: str, namespace: str):
        start = self.NAMESPACE_START + namespace
        if not row.startswith(start):
            return

        vars_str = row[len(start) :]
        if len(vars_str) == 0:
            return

        vars_list = vars_str.split(",")
        vars_tuple = (self.get_var_tuple(var) for var in vars_list)
        vars_dict = {var[0]: var[1] for var in vars_tuple if var}
        return vars_dict

    def get_file_vars(self, file: str, namespace: str):
        result: Dict[str, str] = {}
        with open(file, "r") as f:
            for row in f:
                row_vars = self.get_row_vars(row, namespace)
                if row_vars:
                    result = {**result, **row_vars}

            return result

    def get_namespace_vars(self, namespace: str):
        result: Dict[str, str] = {}
        for file in self.files:
            file_vars = self.get_file_vars(file, namespace)
            if file_vars:
                result = {**result, **file_vars}

        self.var = result
        return result

    def get(self, key: str, default: Optional[str] = None):
        if not self.vars:
            return default

        try:
            value = self.vars[key]
        except KeyError:
            value = default

        return value


def get_options(files: Tuple[str, ...]):
    hledger_vars = HledgerVars(files)
    vars = hledger_vars.get_namespace_vars(NAMESPACE)
    existing_keys = set(vars.keys())
    missing = KEYS.difference(existing_keys)

    if len(missing) > 0:
        raise OptionError(f"Missing keys {missing}")

    errors = ""
    if vars["avg_cost"] not in ["true", "false"]:
        errors += 'avg_cost should be "true" or "false"\n'

    if vars["check"] not in ["true", "false"]:
        errors += 'check should be "true" or "false"\n'

    if errors != "":
        raise OptionError(errors)

    avg_cost = True if vars["avg_cost"] == "true" else False
    check = True if vars["check"] == "true" else False
    no_desc = vars["no_desc"]
    return Options(avg_cost, check, no_desc)



from typing import Optional, Tuple

from .base_args import BaseArgs


class BatchArgs(BaseArgs):
    def __init__(
        self,
        files: Tuple[str, ...],
        name: Optional[str],
        hledger_options: Tuple[str, str],
    ) -> None:
        super().__init__(files)
        self.name = name
        self.hledger_options = hledger_options
        self.report = self.run_batch()

    @property
    def available_txt(self):
        text = "Available args\n\n"

        text += "No interactive allowed args (Report)\n"
        text += "----------------------------\n"
        no_ask_str = "\n".join(self.no_ask)
        text += no_ask_str + "\n\n"

        if len(self.has_ask) > 0:
            text += "Interactive-only args (Report)\n"
            text += "------------------------------\n"
            has_ask_str = "\n".join(self.has_ask)
            text += has_ask_str + "\n\n"
        return text

    def run_one_for_all(self, name: str, hledger_options: Tuple[str, ...]):
        report = self.run_args(name, hledger_options)
        result = f"""================================

Report: {name}

{report}

"""
        return result

    def run_all(self):
        reports = [
            self.run_one_for_all(self.args[name], self.hledger_options)
            for name in self.no_ask
        ]
        reports_txt = "\n".join(reports) + "================================"
        return reports_txt

    def run_batch(self):
        if not self.name:
            return f"Missing command\n\n{self.available_txt}"
        elif self.name == "all":
            return self.run_all()
        elif self.name not in self.names:
            raise KeyError(f"{self.name} not found.\n\n{self.available_txt}")
        elif self.name in self.has_ask:
            raise KeyError(f"{self.name} is interactive only.\n\n{self.available_txt}")
        else:
            options = self.args[self.name]
            return self.run_args(options, self.hledger_options)
